""" This program is for tracking zebra fish around in a tank."""

"""IMPORTS"""
import numpy as np
import cv2
from skimage import feature
import time
import matplotlib.animation as am
from matplotlib import pyplot
import csv
import configparser

"""GLOBALS
    unfortunatly the trajectory must be global due to the way the animator is implemented
"""
trajectory = []

"""classes"""
class Config:
    """The magic numbers for the particle tracking, and other data """
    file_name:str
    output_name:str
    output_traj:str
    min_sigma:float
    max_sigma:float
    threshold:float
    overlap:float
    max_move:int
    backgroundFrame:int
    startFrame:int

    def __init__(self,path):
        config = configparser.ConfigParser()
        config.read(path)
        self.file_name = config['IO']['Input']
        self.output_name = config['IO']['OutputVid']
        self.output_traj = config['IO']['OutputTrajectory']
        self.min_sigma = float(config['Detector']['MinSigma'])
        self.max_sigma = float(config['Detector']['MaxSigma'])
        self. threshold = float(config['Detector']['Threshold'])
        self.overlap = float(config['Detector']['Overlap'])
        self.max_move = int(config['Detector']['MaxMove'])
        self.backgroundFrame = int(config['Background']['BackgroundFrame'])
        self.startFrame = int(config['Background']['StartFrame'])

"""functions"""
def find_fish(blobs,config):
    """find which blob is most likly to be the fish
            Currently looks for the closest blob to the last fish

            The first frame must contain only the Fish, use the
                zerbratest.py to test
    """

    #if no blobs, hold position
    if len(blobs) == 0 and len(trajectory) == 0:
        raise Exception("No objects detetcted in first frame. Please chose another starting frame")
    elif len(blobs) == 0:
        trajectory.append(trajectory[-1])
    elif len(trajectory) == 0:
        trajectory.append(blobs[0])
    else:
        last = trajectory[-1]
        best = 1000000000.0 #(large no)
        best_index = 0
        for i,blob in enumerate(blobs):
            distance = (blob[0]-last[0])*(blob[0]-last[0]) + (blob[1]-last[1])*(blob[1]-last[1])
            if distance < best:
                best = distance
                best_index = i
        if best < config.max_move:
            trajectory.append(blobs[best_index])
        else: trajectory.append(trajectory[-1])


def frameGen(config,background):
    """ generate and process the image frame
        sets the image to greyscale and subtracts the background
    """
    cap = cv2.VideoCapture(config.file_name)
    frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT)) - 1

    for frame_num in range(config.startFrame,frame_count):
        print(frame_num)
        cap.set(1,frame_num)
        ret,frame = cap.read()

        grey = cv2.cvtColor(frame,cv2.COLOR_BGR2GRAY)
        grey = cv2.subtract(background,grey)
        yield  grey
    cap.release()

def step(frame,config,plt,image):
    """Core simulation step, passed to the animtor. Detection is done here"""
    #get points
    my_dog = feature.blob_dog(frame,min_sigma=config.min_sigma,
                                    max_sigma=config.max_sigma,
                                    overlap=config.overlap,
                                    threshold=config.threshold)
    find_fish(my_dog,config)
    #print(my_dog)
    image.set_data(frame)
    plt.set_data([trajectory[-1][1]],
                 [trajectory[-1][0]])            
    return image,

def main():
    """-----------main work code. Execution starts here------------"""

    config = Config('config.ini')

    """get first frame to use as background"""
    cap = cv2.VideoCapture(config.file_name)
    cap.set(1,config.backgroundFrame)
    ret,background = cap.read()
    background = cv2.cvtColor(background,cv2.COLOR_BGR2GRAY)
    cap.release()

    """set up pyplot"""
    fig,ax = pyplot.subplots()
    image=pyplot.imshow(background,cmap='gray')
    plt, = pyplot.plot([],[],'o')

    def _next_frame():
        """closure required for animation input"""
        frame_gen = frameGen(config,background)
        for frame in frame_gen:
            yield frame
    
    ani = am.FuncAnimation(fig,step,frames=_next_frame,
                       fargs=(config,plt,image),
                       save_count=1000000, #big number
                       repeat = False)
    ani.save(config.output_name,writer=am.writers['ffmpeg'](fps=30))

    with open(config.output_traj,'w') as f:
        write = csv.writer(f)
        write.writerows(trajectory)

    cv2.destroyAllWindows()

if __name__ == "__main__":
    main()